package org.xbib.net.http.server.application.database;

import java.sql.SQLException;
import org.xbib.jdbc.pool.api.XbibDataSource;
import org.xbib.jdbc.pool.api.configuration.supplier.XbibConnectionFactoryConfigurationSupplier;
import org.xbib.jdbc.pool.api.configuration.supplier.XbibConnectionPoolConfigurationSupplier;
import org.xbib.jdbc.pool.api.configuration.supplier.XbibDataSourceConfigurationSupplier;
import org.xbib.jdbc.pool.api.security.NamePrincipal;
import org.xbib.jdbc.pool.api.security.SimplePassword;
import org.xbib.jdbc.query.DatabaseProvider;
import org.xbib.net.http.server.application.Application;
import org.xbib.net.http.server.application.BaseApplicationModule;
import org.xbib.net.http.server.HttpRequest;
import org.xbib.net.http.server.route.HttpRouterContext;
import org.xbib.net.http.server.service.HttpService;

import javax.sql.DataSource;
import java.util.Properties;

public class DatabaseApplicationModule extends BaseApplicationModule {

    public DatabaseApplicationModule(Application application, String name) {
        super(application, name);
    }

    @Override
    public void onOpen(HttpRouterContext httpRouterContext, HttpService httpService, HttpRequest httpRequest) {
        try {
            DataSource dataSource = createDataSource();
            String flavor = System.getProperty("database.flavor");
            DatabaseProvider databaseProvider = flavor != null ?
                    DatabaseProvider.builder(dataSource, flavor).build() : null;
            if (dataSource != null) {
                httpRouterContext.getAttributes().put("datasource", dataSource);
            }
            if (databaseProvider != null) {
                httpRouterContext.getAttributes().put("databaseprovider", databaseProvider);
            }
        } catch (SQLException e) {
            throw new IllegalStateException("no datasource can be created");
        }
    }

    private DataSource createDataSource() throws SQLException {
        Properties properties = new Properties();
        System.getProperties().forEach((key, value) -> {
            if (key.toString().startsWith("database.") && !key.toString().equals("database.flavor")) {
                properties.setProperty(key.toString().substring(9), value.toString());
            }
        });
        if (!properties.containsKey("url")) {
            throw new IllegalArgumentException("no database.url in system properties given");
        }
        XbibDataSourceConfigurationSupplier supplier = new XbibDataSourceConfigurationSupplier();
        XbibConnectionPoolConfigurationSupplier poolConfig = supplier.connectionPoolConfiguration();
        poolConfig.maxSize((Integer) properties.getOrDefault("database.poolsize", 4));
        XbibConnectionFactoryConfigurationSupplier connectionFactoryConfig = poolConfig.connectionFactoryConfiguration();
        connectionFactoryConfig.jdbcUrl((String) properties.get("url"));
        if (properties.containsKey("username")) {
            connectionFactoryConfig.credential(new NamePrincipal((String) properties.get("username")));
        }
        if (properties.containsKey("password")) {
            connectionFactoryConfig.credential(new SimplePassword((String) properties.get("password")));
        }
        return XbibDataSource.from(supplier.get());
    }

    private int getAsInt(Properties properties, String key, int defaultValue) {
        if (properties.containsKey(key)) {
            return Integer.parseInt(properties.getProperty(key));
        } else {
            return defaultValue;
        }
    }

    private long getAsLong(Properties properties, String key, long defaultValue) {
        if (properties.containsKey(key)) {
            return Long.parseLong(properties.getProperty(key));
        } else {
            return defaultValue;
        }
    }

    private boolean getAsBoolean(Properties properties, String key, boolean defaultValue) {
        if (properties.containsKey(key)) {
            return Boolean.parseBoolean(properties.getProperty(key));
        } else {
            return defaultValue;
        }
    }
}
