package org.xbib.net.http.server.netty;

import io.netty.buffer.ByteBufAllocator;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.ServerSocketChannel;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import org.xbib.net.http.server.application.Application;
import org.xbib.net.http.server.HttpServerBuilder;
import org.xbib.net.util.NamedThreadFactory;

import java.util.ServiceLoader;
import java.util.concurrent.ThreadFactory;

public class NettyHttpServerBuilder implements HttpServerBuilder {

    private static final Object lock = new Object();

    NettyHttpServerConfig nettyHttpServerConfig;

    Application application;

    ByteBufAllocator byteBufAllocator;

    EventLoopGroup parentEventLoopGroup;

    EventLoopGroup childEventLoopGroup;

    Class<? extends ServerSocketChannel> socketChannelClass;

    HttpChannelInitializer httpChannelInitializer;

    NettyCustomizer nettyCustomizer;

    String secureSocketProviderName;

    protected NettyHttpServerBuilder() {
        this.byteBufAllocator = ByteBufAllocator.DEFAULT;
        this.nettyHttpServerConfig = new NettyHttpServerConfig();
    }

    public NettyHttpServerBuilder setHttpServerConfig(NettyHttpServerConfig nettyHttpServerConfig) {
        this.nettyHttpServerConfig = nettyHttpServerConfig;
        return this;
    }

    public NettyHttpServerBuilder setByteBufAllocator(ByteBufAllocator byteBufAllocator) {
        this.byteBufAllocator = byteBufAllocator;
        return this;
    }

    public NettyHttpServerBuilder setParentEventLoopGroup(EventLoopGroup parentEventLoopGroup) {
        this.parentEventLoopGroup = parentEventLoopGroup;
        return this;
    }

    public NettyHttpServerBuilder setChildEventLoopGroup(EventLoopGroup childEventLoopGroup) {
        this.childEventLoopGroup = childEventLoopGroup;
        return this;
    }

    public NettyHttpServerBuilder setChannelClass(Class<? extends ServerSocketChannel> socketChannelClass) {
        this.socketChannelClass = socketChannelClass;
        return this;
    }

    public NettyHttpServerBuilder setNettyCustomizer(NettyCustomizer nettyCustomizer) {
        this.nettyCustomizer = nettyCustomizer;
        return this;
    }

    public NettyHttpServerBuilder setHttpChannelInitializer(HttpChannelInitializer httpChannelInitializer) {
        this.httpChannelInitializer = httpChannelInitializer;
        return this;
    }

    public NettyHttpServerBuilder setSecureSocketProviderName(String name) {
        this.secureSocketProviderName = name;
        return this;
    }

    public NettyHttpServerBuilder setApplication(Application application) {
        this.application = application;
        return this;
    }

    public NettyHttpServer build() {
        createParentEventLoopGroup(nettyHttpServerConfig);
        createChildEventLoopGroup(nettyHttpServerConfig);
        createSocketChannelClass(nettyHttpServerConfig);
        return new NettyHttpServer(this);
    }

    private void createParentEventLoopGroup(NettyHttpServerConfig httpServerConfig) {
        if (parentEventLoopGroup != null) {
            return;
        }
        synchronized (lock) {
            ThreadFactory threadFactory = new NamedThreadFactory("org-xbib-net-http-netty-server-parent");
            for (ServerTransportProvider serverTransportProvider : ServiceLoader.load(ServerTransportProvider.class)) {
                if (httpServerConfig.getTransportProviderName() == null ||
                        httpServerConfig.getTransportProviderName().equals(serverTransportProvider.getClass().getName())) {
                    parentEventLoopGroup = serverTransportProvider.createEventLoopGroup(httpServerConfig.getParentThreadCount(),
                            threadFactory);
                }
            }
            if (parentEventLoopGroup == null) {
                parentEventLoopGroup = new NioEventLoopGroup(httpServerConfig.getParentThreadCount(), threadFactory);
            }
        }
    }

    private void createChildEventLoopGroup(NettyHttpServerConfig httpServerConfig) {
        if (childEventLoopGroup != null) {
            return;
        }
        synchronized (this) {
            ThreadFactory threadFactory = new NamedThreadFactory("org-xbib-net-http-netty-server-child");
            ServiceLoader<ServerTransportProvider> transportProviders = ServiceLoader.load(ServerTransportProvider.class);
            for (ServerTransportProvider serverTransportProvider : transportProviders) {
                if (httpServerConfig.getTransportProviderName() == null ||
                        httpServerConfig.getTransportProviderName().equals(serverTransportProvider.getClass().getName())) {
                    childEventLoopGroup = serverTransportProvider.createEventLoopGroup(httpServerConfig.getChildThreadCount(),
                            threadFactory);
                }
            }
            if (childEventLoopGroup == null) {
                childEventLoopGroup = new NioEventLoopGroup(httpServerConfig.getChildThreadCount(), threadFactory);
            }
        }
    }

    private void createSocketChannelClass(NettyHttpServerConfig httpServerConfig) {
        if (socketChannelClass != null) {
            return;
        }
        synchronized (lock) {
            ServiceLoader<ServerTransportProvider> transportProviders = ServiceLoader.load(ServerTransportProvider.class);
            for (ServerTransportProvider serverTransportProvider : transportProviders) {
                if (httpServerConfig.getTransportProviderName() == null || httpServerConfig.getTransportProviderName().equals(serverTransportProvider.getClass().getName())) {
                    socketChannelClass = serverTransportProvider.createServerSocketChannelClass();
                    break;
                }
            }
            if (socketChannelClass == null) {
                socketChannelClass = NioServerSocketChannel.class;
            }
        }
    }
}
