package org.xbib.net.http.server.application;

import java.nio.file.Path;
import java.nio.file.Paths;
import java.time.ZoneId;
import java.util.Locale;
import java.util.Objects;
import java.util.Set;

import org.xbib.net.http.server.domain.HttpSecurityDomain;
import org.xbib.net.http.server.executor.BaseExecutor;
import org.xbib.net.http.server.executor.Executor;
import org.xbib.net.http.server.route.HttpRouter;
import org.xbib.net.mime.MimeTypeService;

public class BaseApplicationBuilder implements ApplicationBuilder {

    private static final Set<String> DEFAULT_SUFFIXES =
            Set.of("css", "js", "ico", "png", "jpg", "jpeg", "gif", "woff2");

    protected String name;

    protected String profile;

    protected Path home;

    protected String prefix;

    protected ClassLoader classLoader;

    protected String contextPath;

    protected String secret;

    protected HttpSecurityDomain securityDomain;

    protected boolean sessionsEnabled;

    protected boolean shouldNegotiateLocale;

    protected Locale locale;

    protected ZoneId zoneId;

    protected MimeTypeService mimeTypeService;

    protected Set<String> staticFileSuffixes;

    protected Executor executor;

    protected HttpRouter httpRouter;

    protected BaseApplicationBuilder() {
        this.sessionsEnabled = true;
    }

    public BaseApplicationBuilder setName(String name) {
        this.name = name;
        return this;
    }

    public BaseApplicationBuilder setProfile(String profile) {
        this.profile = profile;
        return this;
    }

    @Override
    public BaseApplicationBuilder setHome(Path home) {
        this.home = home;
        return this;
    }

    @Override
    public BaseApplicationBuilder setPrefix(String prefix) {
        this.prefix = prefix;
        return this;
    }

    public BaseApplicationBuilder setClassLoader(ClassLoader classLoader) {
        this.classLoader = classLoader;
        return this;
    }

    @Override
    public BaseApplicationBuilder setContextPath(String contextPath) {
        this.contextPath = contextPath;
        return this;
    }

    @Override
    public ApplicationBuilder setSecret(String secret) {
        this.secret = secret;
        return this;
    }

    @Override
    public ApplicationBuilder setSecurityDomain(HttpSecurityDomain securityDomain) {
        this.securityDomain = securityDomain;
        return this;
    }

    @Override
    public ApplicationBuilder setSessionsEnabled(boolean sessionsEnabled) {
        this.sessionsEnabled = sessionsEnabled;
        return this;
    }

    @Override
    public ApplicationBuilder shouldNegotiateLocale(boolean shouldNegotiateLocale) {
        this.shouldNegotiateLocale = shouldNegotiateLocale;
        return this;
    }

    @Override
    public ApplicationBuilder setLocale(Locale locale) {
        this.locale = locale;
        return this;
    }

    @Override
    public ApplicationBuilder setZoneId(ZoneId zoneId) {
        this.zoneId = zoneId;
        return this;
    }

    @Override
    public ApplicationBuilder setMimeTypeService(MimeTypeService mimeTypeService) {
        this.mimeTypeService = mimeTypeService;
        return this;
    }

    @Override
    public ApplicationBuilder setStaticSuffixes(String... suffixes) {
        this.staticFileSuffixes = Set.of(suffixes);
        return this;
    }

    @Override
    public ApplicationBuilder setExecutor(Executor executor) {
        this.executor = executor;
        return this;
    }

    @Override
    public ApplicationBuilder setRouter(HttpRouter httpRouter) {
        this.httpRouter = httpRouter;
        return this;
    }

    @Override
    public Application build() {
        Objects.requireNonNull(httpRouter, "http router must not be null");
        if (this.name == null) {
            this.name = System.getProperty("application.name", "application");
        }
        if (this.profile == null) {
            this.profile = System.getProperty("application.profile", "developer");
        }
        if (this.home == null) {
            this.home = Paths.get(System.getProperty("application.home", "."));
        }
        if (this.prefix == null) {
            this.prefix = "/";
        }
        if (this.classLoader == null) {
            this.classLoader = getClass().getClassLoader();
        }
        if (this.contextPath == null) {
            this.contextPath = "/";
        }
        if (this.secret == null) {
            this.secret = "secret";
        }
        if (this.locale == null) {
            this.locale = Locale.getDefault();
        }
        if (this.zoneId == null) {
            this.zoneId = ZoneId.systemDefault();
        }
        if (this.mimeTypeService == null) {
            this.mimeTypeService = new MimeTypeService();
        }
        if (this.staticFileSuffixes == null) {
            this.staticFileSuffixes = DEFAULT_SUFFIXES;
        }
        if (executor == null) {
            this.executor = BaseExecutor.builder().build();
        }
        /*for (Map.Entry<String, Settings> entry : settings.getGroups("module").entrySet()) {
            String moduleName = entry.getKey();
            Settings moduleSettings = entry.getValue();
            if (moduleSettings.getAsBoolean("enabled", true)) {
                try {
                    String className = moduleSettings.get("class");
                    @SuppressWarnings("unchecked")
                    Class<ApplicationModule> clazz =
                            (Class<ApplicationModule>) Class.forName(className, true, builder.classLoader);
                    ApplicationModule applicationModule = clazz.getConstructor(Application.class, String.class, Settings.class)
                            .newInstance(this, moduleName, moduleSettings);
                    applicationModuleList.add(applicationModule);
                } catch (Exception e) {
                    logger.log(Level.WARNING, e.getMessage(), e);
                    throw new IllegalArgumentException("class not found or not loadable: " + e.getMessage());
                }
            } else {
                logger.log(Level.WARNING, "disabled module: " + moduleName);
            }
        }*/
        return new BaseApplication(this);
    }
}
