package org.xbib.net.http.server.auth;

import org.xbib.net.PercentEncoder;
import org.xbib.net.PercentEncoders;
import org.xbib.net.UserProfile;
import org.xbib.net.http.server.persist.Codec;
import org.xbib.net.util.JsonUtil;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.io.Writer;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.attribute.FileTime;
import java.time.Duration;
import java.time.Instant;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Stream;

public class FileJsonUserProfileCodec implements Codec<UserProfile> {

    private static final Logger logger = Logger.getLogger(FileJsonUserProfileCodec.class.getName());

    private final ReentrantReadWriteLock lock;

    private final Path path;

    public FileJsonUserProfileCodec(Path path) {
        this.path = path;
        this.lock = new ReentrantReadWriteLock();
    }

    @Override
    public UserProfile create(String userId) throws IOException {
        Objects.requireNonNull(userId, "user id must not be null");
        BaseUserProfile baseUserProfile = new BaseUserProfile();
        baseUserProfile.setUserId(userId);
        return baseUserProfile;
    }

    @Override
    public UserProfile read(String key) throws IOException {
        Objects.requireNonNull(key, "key must not be null");
        ReentrantReadWriteLock.ReadLock readLock = lock.readLock();
        try {
            readLock.lock();
            PercentEncoder percentEncoder = PercentEncoders.getUnreservedEncoder(StandardCharsets.UTF_8);
            Path p = path.resolve(percentEncoder.encode(key));
            Map<String, Object> map = JsonUtil.toMap(Files.readString(p));
            return BaseUserProfile.fromMap(map);
        } finally {
            readLock.unlock();
        }
    }

    @Override
    public void write(String key, UserProfile userProfile) throws IOException {
        Objects.requireNonNull(key, "key must not be null");
        ReentrantReadWriteLock.WriteLock writeLock = lock.writeLock();
        try {
            writeLock.lock();
            createPath();
            PercentEncoder percentEncoder = PercentEncoders.getUnreservedEncoder(StandardCharsets.UTF_8);
            try (Writer writer = Files.newBufferedWriter(path.resolve(percentEncoder.encode(key)))) {
                writer.write(JsonUtil.toString(userProfile.asMap()));
            }
        } finally {
            writeLock.unlock();
        }
    }

    @Override
    public void remove(String key) {
        ReentrantReadWriteLock.WriteLock writeLock = lock.writeLock();
        try {
            writeLock.lock();
            PercentEncoder percentEncoder = PercentEncoders.getUnreservedEncoder(StandardCharsets.UTF_8);
            Files.deleteIfExists(path.resolve(percentEncoder.encode(key)));
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        } finally {
            writeLock.unlock();
        }
    }

    @Override
    public void purge(long expiredAfterSeconds) {
        if (path != null && expiredAfterSeconds > 0L) {
            Instant instant = Instant.now();
            try (Stream<Path> stream = Files.walk(path)) {
                stream.forEach(p -> {
                    try {
                        FileTime fileTime = Files.getLastModifiedTime(p);
                        Duration duration = Duration.between(fileTime.toInstant(), instant);
                        if (duration.toSeconds() > expiredAfterSeconds) {
                            Files.delete(p);
                        }
                    } catch (IOException e) {
                        logger.log(Level.WARNING, "i/o error while purge: " + e.getMessage(), e);
                    }
                });
            } catch (IOException e) {
                logger.log(Level.WARNING, "i/o error while purge: " + e.getMessage(), e);
            }
        }
    }

    @Override
    public void destroy() {
        try (Stream<Path> stream = Files.walk(path)) {
            stream.forEach(p -> {
                try {
                    if (!path.equals(p)) {
                        Files.delete(p);
                    }
                } catch (IOException e) {
                    logger.log(Level.WARNING, "i/o error while destroy: " + e.getMessage(), e);
                }
            });
        } catch (IOException e) {
            logger.log(Level.WARNING, "i/o error while destroy: " + e.getMessage(), e);
        }
        try {
            Files.delete(path);
        } catch (IOException e) {
            logger.log(Level.WARNING, "i/o error while destroy: " + e.getMessage(), e);
        }
    }

    private void createPath() {
        try {
            if (!Files.exists(path)) {
                Files.createDirectories(path);
            }
        } catch (IOException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            throw new UncheckedIOException(e);
        }
    }
}
