package org.xbib.net.http.server.auth;

import java.io.IOException;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.net.Parameter;
import org.xbib.net.ParameterException;
import org.xbib.net.SecurityRealm;
import org.xbib.net.URL;
import org.xbib.net.UserProfile;
import org.xbib.net.http.server.HttpHandler;
import org.xbib.net.http.server.route.HttpRouterContext;

public class FormAuthenticationHandler extends LoginAuthenticationHandler implements HttpHandler {

    private static final Logger logger = Logger.getLogger(FormAuthenticationHandler.class.getName());

    public FormAuthenticationHandler(String usernameParameter,
                                     String passwordParameter,
                                     SecurityRealm securityRealm) {
        super(usernameParameter, passwordParameter, securityRealm);
    }

    @Override
    public void handle(HttpRouterContext context) throws IOException {
        logger.log(Level.FINE, "request = " + context.getRequest());
        UserProfile userProfile = context.getAttributes().get(UserProfile.class, "userprofile");
        if (userProfile != null && userProfile.getUserId() != null) {
            logger.log(Level.FINE, "already authenticated, user id = " + userProfile.getUserId());
            context.setAuthenticated(true);
            return;
        }
        if (context.isAuthenticated()) {
            logger.log(Level.FINE, "context already authenticated");
            return;
        }
        // always add an "anonymous" user profile with a null user ID
        userProfile = new BaseUserProfile();
        boolean isAuthenticated = false;
        try {
            Parameter parameter = context.getRequest().getParameter();
            if (!parameter.containsKey(getUserParameterName(), Parameter.Domain.FORM)) {
                logger.log(Level.WARNING, "username parameter not set, unable to authenticate");
                prepareFormAuthentication(context);
                return;
            }
            String username = parameter.getAsString(getUserParameterName(), Parameter.Domain.FORM);
            if (!parameter.containsKey(getPasswordParameterName(), Parameter.Domain.FORM)) {
                logger.log(Level.WARNING, "password parameter not set, unable to authenticate");
                prepareFormAuthentication(context);
                return;
            }
            String password = parameter.getAsString(getPasswordParameterName(), Parameter.Domain.FORM);
            if (authenticate(userProfile, username, password, context.getRequest())) {
                isAuthenticated = true;
            }
        } catch (ParameterException e) {
            logger.log(Level.SEVERE, "parameter error");
        } catch (Exception e) {
            logger.log(Level.SEVERE, "authentication error");
        } finally {
            context.getAttributes().put("userprofile", userProfile);
            context.setAuthenticated(isAuthenticated);
            if (!isAuthenticated) {
                prepareFormAuthentication(context);
            }
        }
    }

    protected void prepareFormAuthentication(HttpRouterContext context) {
        URL origin = context.getContextURL().resolve(context.getRequest().getRequestURI()).normalize();
        logger.log(Level.FINE, "context URL = " + context.getContextURL() +
                " request URI = " + context.getRequest().getRequestURI() +
                " origin = " + origin);
        context.getAttributes().put("_origin", origin.toExternalForm());
    }
}
