package org.xbib.net.http.server.auth;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.net.Authenticator;
import org.xbib.net.GroupsProvider;
import org.xbib.net.Parameter;
import org.xbib.net.ParameterException;
import org.xbib.net.Request;
import org.xbib.net.SecurityRealm;
import org.xbib.net.UserDetails;
import org.xbib.net.UserProfile;
import org.xbib.net.UsersProvider;
import org.xbib.net.http.server.HttpHandler;
import org.xbib.net.http.server.route.HttpRouterContext;

public class LoginAuthenticationHandler implements HttpHandler {

    private static final Logger logger = Logger.getLogger(LoginAuthenticationHandler.class.getName());

    private final String userParameterName;

    private final String passwordParameterName;

    private final SecurityRealm securityRealm;

    public LoginAuthenticationHandler(String usernameParameterName,
                                      String passwordParameterName,
                                      SecurityRealm securityRealm) {
        this.userParameterName = Objects.requireNonNull(usernameParameterName, "username parameter must not be null");
        this.passwordParameterName = Objects.requireNonNull(passwordParameterName, "password parameter must not be null");
        this.securityRealm = Objects.requireNonNull(securityRealm, "security realm must not be null");
    }

    public String getUserParameterName() {
        return userParameterName;
    }

    public String getPasswordParameterName() {
        return passwordParameterName;
    }

    public SecurityRealm getSecurityRealm() {
        return securityRealm;
    }

    @Override
    public void handle(HttpRouterContext context) throws IOException {
        UserProfile userProfile = context.getAttributes().get(UserProfile.class, "userprofile");
        if (userProfile == null) {
            logger.log(Level.FINE, "no user profile found in context");
            return;
        }
        if (userProfile.isRefreshTriggered()) {
            try {
                List<String> roles = new ArrayList<>();
                userProfile.setRoles(roles);
                List<String> effectiveRoles = new ArrayList<>();
                userProfile.setEffectiveRoles(effectiveRoles);
                loadRoles(userProfile);
            } catch (Exception e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            } finally {
                userProfile.triggerRefresh(false);
            }
            return;
        }
        if (userProfile.getUserId() != null) {
            logger.log(Level.FINE, "user id already set");
            return;
        }
        Parameter parameter = context.getRequest().getParameter();
        userProfile = new BaseUserProfile();
        context.getAttributes().put("userprofile", userProfile);
        boolean isAuthenticated = false;
        try {
            String username = parameter.getAsString(getUserParameterName(), Parameter.Domain.FORM);
            String password = parameter.getAsString(getPasswordParameterName(), Parameter.Domain.FORM);
            if (authenticate(userProfile, username, password, context.getRequest())) {
                isAuthenticated = true;
            }
        } catch (ParameterException e) {
            logger.log(Level.SEVERE, "parameter error");
        } catch (Exception e) {
            logger.log(Level.SEVERE, "authentication error");
        } finally {
            context.setAuthenticated(isAuthenticated);
        }
    }

    protected boolean authenticate(UserProfile userProfile,
                                   String username,
                                   String password,
                                   Request request) {
        if (username == null) {
            logger.log(Level.FINE, "no username given for check, doing nothing");
            return false;
        }
        if (password == null) {
            logger.log(Level.FINE, "no password given for check, doing nothing");
            return false;
        }
        Authenticator auth = securityRealm.getAuthenticator();
        Authenticator.Context authContext = new Authenticator.Context(username, password, request);
        if (auth.authenticate(authContext)) {
            logger.log(Level.FINE, "authenticated as " + authContext.getUsername());
            userProfile.setUserId(authContext.getUsername());
            UsersProvider.Context userContext = new UsersProvider.Context(authContext.getUsername(), null);
            UserDetails userDetails = securityRealm.getUsersProvider().getUserDetails(userContext);
            userProfile.setName(userDetails.getName());
            if (userDetails.getEffectiveUserId() != null) {
                userProfile.setEffectiveUserId(userDetails.getEffectiveUserId());
            } else {
                userProfile.setEffectiveUserId(authContext.getUsername());
            }
            loadRoles(userProfile);
            return true;
        }
        logger.log(Level.FINE, "authentication failure for user " + username);
        return false;
    }

    private void loadRoles(UserProfile userProfile) {
        GroupsProvider.Context groupContext = new GroupsProvider.Context(userProfile.getUserId(), null);
        Collection<String> groups = securityRealm.getGroupsProvider().getGroups(groupContext);
        for (String group : groups) {
            userProfile.addRole(group);
        }
        if (!userProfile.getUserId().equals(userProfile.getEffectiveUserId())) {
            GroupsProvider.Context effectiveGroupContext = new GroupsProvider.Context(userProfile.getEffectiveUserId(), null);
            for (String group : securityRealm.getGroupsProvider().getGroups(effectiveGroupContext)) {
                userProfile.addEffectiveRole(group);
            }
        } else {
            for (String group : groups) {
                userProfile.addEffectiveRole(group);
            }
        }
    }
}
