package org.xbib.net.http.server.auth;

import org.xbib.net.GroupsProvider;
import org.xbib.net.SecurityRealm;
import org.xbib.net.UserProfile;
import org.xbib.net.http.HttpResponseStatus;
import org.xbib.net.http.server.HttpException;
import org.xbib.net.http.server.HttpHandler;
import org.xbib.net.http.server.domain.HttpSecurityDomain;
import org.xbib.net.http.server.persist.Codec;
import org.xbib.net.http.server.route.HttpRouterContext;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

public class PersistUserProfileHandler implements HttpHandler {

    private static final Logger logger = Logger.getLogger(PersistUserProfileHandler.class.getName());

    private final Codec<UserProfile> userProfileCodec;

    private final HttpSecurityDomain securityDomain;

    public PersistUserProfileHandler(Codec<UserProfile> userProfileCodec,
                                     HttpSecurityDomain securityDomain) {
        this.userProfileCodec = userProfileCodec;
        this.securityDomain = securityDomain;
    }

    @Override
    public void handle(HttpRouterContext context) throws IOException {
        UserProfile userProfile = context.getAttributes().get(UserProfile.class, "userprofile");
        if (userProfile != null && userProfile.getUserId() != null) {
            if (userProfile.isRefreshTriggered()) {
                if (securityDomain == null) {
                    logger.log(Level.FINE, "unable to refresh, security domain is null");
                    return;
                }
                if (securityDomain.getRealm() == null) {
                    logger.log(Level.FINE, "unable to refresh, security realm is null");
                    return;
                }
                try {
                    // reset roles and load again
                    List<String> roles = new ArrayList<>();
                    userProfile.setRoles(roles);
                    List<String> effectiveRoles = new ArrayList<>();
                    userProfile.setEffectiveRoles(effectiveRoles);
                    loadRoles(userProfile, securityDomain.getRealm());
                } catch (Exception e) {
                    logger.log(Level.SEVERE, e.getMessage(), e);
                } finally {
                    userProfile.triggerRefresh(false);
                }
            }
            try {
                logger.log(Level.FINEST, "writing user profile id  = " + userProfile.getUserId());
                userProfileCodec.write(userProfile.getUserId(), userProfile);
            } catch (Exception e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
                throw new HttpException("unable to write user profile data", context, HttpResponseStatus.INTERNAL_SERVER_ERROR);
            }
        } else {
            logger.log(Level.FINEST, "not writing user profile " + userProfile + " user id + " +
                    (userProfile != null ? userProfile.getUserId() : null));
        }
    }

    private void loadRoles(UserProfile userProfile, SecurityRealm securityRealm) {
        String username = userProfile.getUserId();
        GroupsProvider.Context groupContext = new GroupsProvider.Context(username, null);
        Collection<String> groups = securityRealm.getGroupsProvider().getGroups(groupContext);
        for (String group : groups) {
            userProfile.addRole(group);
        }
        if (!userProfile.getUserId().equals(userProfile.getEffectiveUserId())) {
            GroupsProvider.Context effectiveGroupContext = new GroupsProvider.Context(userProfile.getEffectiveUserId(), null);
            for (String group : securityRealm.getGroupsProvider().getGroups(effectiveGroupContext)) {
                userProfile.addEffectiveRole(group);
            }
        } else {
            for (String group : groups) {
                userProfile.addEffectiveRole(group);
            }
        }
    }
}
