package org.xbib.net.http.server.route;

import java.security.KeyStoreException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xbib.net.URL;
import org.xbib.net.http.HttpAddress;
import org.xbib.net.http.HttpMethod;
import org.xbib.net.http.HttpVersion;
import org.xbib.net.http.server.HttpHandler;
import org.xbib.net.http.server.HttpServiceBuilder;
import org.xbib.net.http.server.domain.BaseHttpDomain;
import org.xbib.net.http.server.domain.HttpDomain;
import org.xbib.net.http.server.domain.HttpDomainBuilder;
import org.xbib.net.http.server.domain.HttpSecurityDomain;
import org.xbib.net.http.server.handler.BadRequestHandler;
import org.xbib.net.http.server.handler.ForbiddenHandler;
import org.xbib.net.http.server.handler.InternalServerErrorHandler;
import org.xbib.net.http.server.handler.NotFoundHandler;
import org.xbib.net.http.server.handler.NotImplementedHandler;
import org.xbib.net.http.server.handler.UnauthorizedHandler;
import org.xbib.net.http.server.handler.VersionNotSupportedHandler;
import org.xbib.net.http.server.service.BaseHttpService;
import org.xbib.net.http.server.service.HttpService;
import org.xbib.settings.Settings;

import javax.net.ssl.SSLException;

public class BaseHttpRouterBuilder implements HttpRouterBuilder {

    private static final Logger logger = Logger.getLogger(BaseHttpRouterBuilder.class.getName());

    protected String prefix;

    protected final Collection<HttpDomain> domains;

    protected final Map<Integer, HttpHandler> handlers;

    protected HttpRouteResolver<HttpService> httpRouteResolver;

    protected BaseHttpRouterBuilder() {
        this.prefix = "";
        this.domains = new ArrayList<>();
        this.handlers = new HashMap<>();
        // default handlers
        this.handlers.put(400, new BadRequestHandler());
        this.handlers.put(401, new UnauthorizedHandler());
        this.handlers.put(403, new ForbiddenHandler());
        this.handlers.put(404, new NotFoundHandler());
        this.handlers.put(500, new InternalServerErrorHandler());
        this.handlers.put(501, new NotImplementedHandler());
        this.handlers.put(505, new VersionNotSupportedHandler());
    }

    @Override
    public BaseHttpRouterBuilder setPrefix(String prefix) {
        Objects.requireNonNull(prefix);
        // Add ending slash if missing.
        // We require a prefix with ending slash. Otherwise, obscure things can happen in path parameter handling,
        // if a path parameter has a common prefix with this prefix.
        this.prefix = prefix.isEmpty() || prefix.endsWith("/") ? prefix : prefix + "/";
        return this;
    }

    @Override
    public BaseHttpRouterBuilder setHandler(Integer code, HttpHandler httpHandler) {
        if (httpHandler != null) {
            handlers.put(code, httpHandler);
        }
        return this;
    }

    @Override
    public BaseHttpRouterBuilder addDomain(HttpDomain domain) {
        this.domains.add(domain);
        return this;
    }

    @Override
    public BaseHttpRouterBuilder setRouteResolver(HttpRouteResolver<HttpService> httpRouteResolver) {
        this.httpRouteResolver = httpRouteResolver;
        return this;
    }

    @Override
    public BaseHttpRouter build() {
        if (domains.isEmpty()) {
            throw new IllegalArgumentException("no domain configured, unable to continue");
        }
        if (httpRouteResolver == null) {
            HttpRouteResolver.Builder<HttpService> httpRouteResolverBuilder = BaseHttpRouteResolver.builder();
            for (HttpDomain domain : domains) {
                for (HttpService httpService : domain.getServices()) {
                    logger.log(Level.FINER, "adding " + domain.getAddress() + " " + httpService.getMethods() +
                            " prefix = " + httpService.getPrefix() +
                            " path = " + httpService.getPathSpecification() + " " + httpService);
                    HttpRoute httpRoute = new BaseHttpRoute(domain.getAddress(),
                            httpService.getMethods(),
                            httpService.getPrefix(),
                            httpService.getPathSpecification(),
                            false);
                    httpRouteResolverBuilder.add(httpRoute, httpService);
                }
            }
            this.httpRouteResolver = httpRouteResolverBuilder.build();
        }
        return new BaseHttpRouter(this);
    }


    public HttpRouterBuilder fromSettings(Settings settings) {
        fromSettings(settings, null);
        return this;
    }

    public HttpRouterBuilder fromSettings(Settings settings,
                                          Map<String, HttpSecurityDomain> securityDomainMap) {
        // global prefix
        if (settings.containsSetting("prefix")) {
            setPrefix(settings.get("prefix"));
        }
        // global handlers
        for (Map.Entry<String, Settings> entry : settings.getGroups("handler").entrySet()) {
            String codeString = entry.getKey();
            Settings handlerSettings = entry.getValue();
            setHandler(Integer.parseInt(codeString), createHttpHandler(handlerSettings));
        }
        // handler domains
        for (Map.Entry<String, Settings> entry : settings.getGroups("domain").entrySet()) {
            addDomain(createHttpDomain(entry.getValue(), securityDomainMap));
        }
        return this;
    }

    private HttpDomain createHttpDomain(Settings settings,
                                        Map<String, HttpSecurityDomain> securityDomainMap) {
        if (settings.getAsBoolean("enabled", true)) {
            try {
                HttpDomainBuilder httpDomainBuilder = BaseHttpDomain.builder();
                httpDomainBuilder.setHttpAddress(createHttpAddress(settings));
                for (String name : settings.getAsArray("name")) {
                    httpDomainBuilder.addName(name);
                }
                for (Map.Entry<String, Settings> entry : settings.getGroups("service").entrySet()) {
                    httpDomainBuilder.addService(createHttpService(securityDomainMap, prefix, entry.getKey(), entry.getValue()));
                }
                return httpDomainBuilder.build();
            } catch (Exception e) {
                logger.log(Level.WARNING, e.getMessage(), e);
                throw new IllegalArgumentException("class not found or not loadable: " + e.getMessage());
            }
        } else {
            logger.log(Level.WARNING, "disabled domain: " + settings.getAsMap());
        }
        return null;
    }

    private HttpAddress createHttpAddress(Settings settings) throws KeyStoreException, SSLException {
        HttpAddress.Builder httpAddressBuilder = HttpAddress.builder();
        if (settings.containsSetting("url")) {
            URL url = URL.create(settings.get("url"));
            httpAddressBuilder.setHost(url.getHost());
            httpAddressBuilder.setPort(url.getPort());
            httpAddressBuilder.setSecure("https".equals(url.getScheme()) || "h2".equals(url.getScheme()));
            httpAddressBuilder.setVersion(url.getScheme().startsWith("http") ? HttpVersion.HTTP_1_1 : HttpVersion.HTTP_2_0);
        } else {
            httpAddressBuilder.setHost(settings.get("host", "localhost"));
            boolean isSecure = settings.getAsBoolean("secure", false);
            httpAddressBuilder.setSecure(isSecure);
            httpAddressBuilder.setPort(settings.getAsInt("port", isSecure ? 443 : 80));
            httpAddressBuilder.setVersion(HttpVersion.valueOf(settings.get("version", "HTTP/1.1")));
        }
        return httpAddressBuilder.build();
    }

    private HttpService createHttpService(Map<String, HttpSecurityDomain> securityDomainMap,
                                          String prefix,
                                          String path,
                                          Settings settings) {
        HttpServiceBuilder httpServiceBuilder = BaseHttpService.builder();
        if (settings.containsSetting("securitydomain")) {
            if (securityDomainMap != null) {
                httpServiceBuilder.setSecurityDomain(securityDomainMap.get(settings.get("securitydomain")));
            }
        }
        httpServiceBuilder.setPrefix(prefix);
        httpServiceBuilder.setPath(path);
        if (settings.containsSetting("method")) {
            String[] methods = settings.getAsArray("method");

        } else {
            httpServiceBuilder.addMethod(HttpMethod.GET);
        }
        httpServiceBuilder.addHandler(createHttpHandler(settings));
        return httpServiceBuilder.build();
    }

    private HttpHandler createHttpHandler(Settings settings) {
        if (settings.getAsBoolean("enabled", true)) {
            try {
                String className = settings.get("class");
                @SuppressWarnings("unchecked")
                Class<HttpHandler> clazz = (Class<HttpHandler>) Class.forName(className, true, getClass().getClassLoader());
                return clazz.getConstructor().newInstance();
            } catch (Exception e) {
                logger.log(Level.WARNING, e.getMessage(), e);
                throw new IllegalArgumentException("class not found or not loadable: " + e.getMessage());
            }
        } else {
            logger.log(Level.WARNING, "disabled: " + settings.getAsMap());
        }
        return null;
    }
}
