package org.xbib.net.http.server.session;

import java.io.IOException;
import java.nio.charset.CodingErrorAction;
import java.nio.charset.StandardCharsets;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.text.MessageFormat;
import java.util.Map;
import java.util.Set;
import java.util.function.Supplier;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.net.PercentDecoder;
import org.xbib.net.UserProfile;
import org.xbib.net.http.HttpResponseStatus;
import org.xbib.net.http.cookie.Cookie;
import org.xbib.net.http.cookie.CookieBox;
import org.xbib.net.http.server.HttpException;
import org.xbib.net.http.server.HttpHandler;
import org.xbib.net.http.server.auth.BaseUserProfile;
import org.xbib.net.http.server.route.HttpRouterContext;
import org.xbib.net.http.server.cookie.CookieSignatureException;
import org.xbib.net.http.server.cookie.CookieSignatureUtil;
import org.xbib.net.http.server.persist.Codec;
import org.xbib.net.util.ExceptionFormatter;

public class IncomingContextHandler implements HttpHandler {

    private static final Logger logger = Logger.getLogger(IncomingContextHandler.class.getName());

    private final Codec<UserProfile> userProfileCodec;

    private final String sessionSecret;

    private final String sessionCookieAlgorithm;

    private final String sessionCookieName;

    private final Codec<Session> sessionCodec;

    private final Set<String> suffixes;

    Supplier<String> sessionIdGenerator;

    public IncomingContextHandler(Codec<UserProfile> userProfileCodec,
                                  String sessionSecret,
                                  String sessionCookieAlgorithm,
                                  String sessionCookieName,
                                  Codec<Session> sessionCodec,
                                  Set<String> suffixes,
                                  Supplier<String> sessionIdGenerator) {
        this.userProfileCodec = userProfileCodec;
        this.sessionSecret = sessionSecret;
        this.sessionCookieAlgorithm = sessionCookieAlgorithm;
        this.sessionCookieName = sessionCookieName;
        this.sessionCodec = sessionCodec;
        this.suffixes = suffixes;
        this.sessionIdGenerator = sessionIdGenerator;
    }

    @Override
    public void handle(HttpRouterContext context) throws HttpException {
        String suffix = SessionUtil.extractExtension(context.getRequestBuilder().getRequestPath());
        if (suffix != null && suffixes.contains(suffix)) {
            return;
        }
        Map<String, Object> payload = null;
        UserProfile userProfile = null;
        Session session = null;
        CookieBox cookieBox = context.getAttributes().get(CookieBox.class, "incomingcookies");
        if (cookieBox != null) {
            for (Cookie cookie : cookieBox) {
                if (cookie.name().equals(sessionCookieName)) {
                    logger.log(Level.FINE, "found our cookie " + sessionCookieName);
                    if (session == null) {
                        try {
                            // extract payload from our cookie, must be not null, otherwise security problem with exception
                            payload = toPayload(cookie);
                            logger.log(Level.FINE, payload != null && !payload.isEmpty() ? "payload found" : "no payload found");
                            // extract session from payload and recover session from persistence store
                            session = toSession(payload);
                            // do not log explicit content of payload, session, userprofile to not leak sensitive info into log
                            logger.log(Level.FINE, session != null && !session.isEmpty() ? "session found" : "no session found");
                            // extract userprofile from cookie info, use previous auth handler setup, recover attributes only
                            userProfile = recoverUserProfile(context, session, payload);
                            logger.log(Level.FINE, userProfile != null ? "user profile found" : "no user profile found");
                        } catch (CookieSignatureException e) {
                            // set exception in context to discard broken cookie later and render exception message
                            context.getAttributes().put("_throwable", e);
                        } catch (Exception e) {
                            logger.log(Level.SEVERE, e.getMessage(), e);
                            throw new HttpException("unable to create userprofile or session: " + ExceptionFormatter.format(e),
                                    context, HttpResponseStatus.INTERNAL_SERVER_ERROR);
                        }
                    } else {
                        logger.log(Level.WARNING, "received extra session cookie of same name, something is wrong, ignoring");
                    }
                }
            }
        }
        if (session == null) {
            session = createSession(context);
            if (session != null) {
                logger.log(Level.FINE, "new session created, id = " + session.id());
            }
        }
        context.getAttributes().put("session", session);
        if (userProfile == null) {
            userProfile = recoverUserProfile(context, session, payload);
        }
        context.getAttributes().put("userprofile", userProfile);
    }

    protected Map<String, Object> toPayload(Cookie cookie) throws IOException,
            NoSuchAlgorithmException, InvalidKeyException, CookieSignatureException {
        PercentDecoder percentDecoder = new PercentDecoder(StandardCharsets.ISO_8859_1.newDecoder()
                .onMalformedInput(CodingErrorAction.IGNORE)
                .onUnmappableCharacter(CodingErrorAction.IGNORE));
        String value = percentDecoder.decode(cookie.value());
        String[] s = value.split(":", 3);
        if (s.length != 3) {
            throw new IllegalArgumentException("cookie format problem, not 3 groups: " + cookie.value());
        }
        String id = s[0];
        String payload = s[1];
        String sig = s[2];
        String mysig = CookieSignatureUtil.hmac(payload, sessionSecret, sessionCookieAlgorithm);
        if (!sig.equals(mysig)) {
            logger.log(Level.SEVERE, () -> MessageFormat.format("signature in cookie does not match. algo={0} secret={1} payload={2} sig={3} mysig={4}",
                    sessionCookieAlgorithm, sessionSecret, payload, sig, mysig));
            throw new CookieSignatureException("cookie security problem");
        }
        return Map.of("id", id, "payload", payload, "map", CookieSignatureUtil.toMap(payload));
    }

    protected Session toSession(Map<String, Object> payload) {
        if (payload == null) {
            return null;
        }
        String id = (String) payload.get("id");
        Session session = null;
        try {
            if (id != null) {
                session = sessionCodec.read(id);
                if (session != null) {
                    logger.log(Level.FINE, "session id " + id + " restored, valid = " + session.isValid() +
                            ", age = " + session.getAge() + ", expired = " + session.isExpired());
                    session.putAll(payload);
                }
            }
        } catch (Exception e) {
            logger.log(Level.FINEST, "unable to restore session id = " + id, e);
        }
        return session;
    }

    protected Session createSession(HttpRouterContext context) throws HttpException {
        try {
            return sessionCodec.create(sessionIdGenerator.get());
        } catch (IOException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            throw new HttpException("unable to create new session", context, HttpResponseStatus.INTERNAL_SERVER_ERROR);
        }
    }

    @SuppressWarnings("unchecked")
    protected UserProfile recoverUserProfile(HttpRouterContext context,
                                             Session session,
                                             Map<String, Object> payload) {
        // already in context from previous handler?
        UserProfile userProfile = context.getAttributes().get(UserProfile.class, "userprofile");
        if (userProfile != null) {
            try {
                return userProfileCodec.read(userProfile.getUserId());
            } catch (IOException e) {
                // ignore I/O error, leads to empty user profile
            }
        }
        // initialize
        Map<String, Object> map = null;
        // is user profile present in session?
        if (payload == null && session != null) {
            // session has stored user profile? this is important for spanning HTTP request/response like in HTTP POST/FORWARD/GET
            map = (Map<String, Object>) session.get("userprofile");
        } else if (payload != null) {
            // is a mini user profile otherwise present in our cookie?
            map = (Map<String, Object>) payload.get("map");
        }
        if (map != null && map.containsKey(UserProfile.USER_ID)) {
            try {
                return userProfileCodec.read(BaseUserProfile.fromMap(map).getUserId());
            } catch (IOException e) {
                // ignore I/O error, leads to incomplete user profile
            }
        } else {
            logger.log(Level.FINE, "unable to find info for initialize user profile");
        }
        return null;
    }
}
