package org.xbib.net.http.template.groovy;

import org.xbib.net.Parameter;
import org.xbib.net.ParameterException;
import org.xbib.net.SecurityRealm;
import org.xbib.net.URL;
import org.xbib.net.UserProfile;
import org.xbib.net.http.server.auth.BaseUserProfile;
import org.xbib.net.http.server.auth.FormAuthenticationHandler;
import org.xbib.net.http.server.route.HttpRouterContext;

import java.io.IOException;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;

public class GroovyFormAuthenticationHandler extends FormAuthenticationHandler {

    private static final Logger logger = Logger.getLogger(GroovyFormAuthenticationHandler.class.getName());

    private final String loginTemplate;

    public GroovyFormAuthenticationHandler(String usernameParameter,
                                           String passwordParameter,
                                           SecurityRealm securityRealm,
                                           String loginTemplate) {
        super(usernameParameter, passwordParameter, securityRealm);
        this.loginTemplate = Objects.requireNonNull(loginTemplate, "no login template configured");
    }

    @Override
    public void handle(HttpRouterContext context) throws IOException {
        UserProfile userProfile = context.getAttributes().get(UserProfile.class, "userprofile");
        if (userProfile != null && userProfile.getUserId() != null) {
            logger.log(Level.FINE, "user id already set: " + userProfile.getUserId());
            return;
        }
        // always add an "anonymous" user profile
        userProfile = new BaseUserProfile();
        context.getAttributes().put("userprofile", userProfile);
        Parameter parameter = context.getRequest().getParameter();
        if (!parameter.containsKey(getUserParameterName(), Parameter.Domain.FORM)) {
            logger.log(Level.WARNING, "usernameParameter not set, unable to authenticate");
            prepareFormAuthentication(context);
            return;
        }
        if (!parameter.containsKey(getPasswordParameterName(), Parameter.Domain.FORM)) {
            logger.log(Level.WARNING, "passwordParameter not set, unable to authenticate");
            prepareFormAuthentication(context);
            return;
        }
        try {
            String username = parameter.getAsString(getUserParameterName(), Parameter.Domain.FORM);
            String password = parameter.getAsString(getPasswordParameterName(), Parameter.Domain.FORM);
            logger.log(Level.FINE, "username and password found, ready for authentication");
            if (authenticate(userProfile, username, password, context.getRequest())) {
                context.setAuthenticated(true);
                logger.log(Level.FINE, "successful authentication");
                return;
            }
        } catch (ParameterException e) {
            logger.log(Level.SEVERE, "parameter error");
        } catch (Exception e) {
            logger.log(Level.SEVERE, "authentication error");
        }
        prepareFormAuthentication(context);
    }

    @Override
    protected void prepareFormAuthentication(HttpRouterContext context) {
        // this will redirect internally to login page, and back to the original path.
        // We need a full path resolve against the server URL.
        logger.log(Level.FINE, "set new templatePath, login template = " + loginTemplate);
        context.getAttributes().put("templatePath", loginTemplate);
        URL origin = context.getContextURL().resolve(context.getRequest().getRequestURI()).normalize();
        logger.log(Level.FINE, "context URL = " + context.getContextURL() + " request URI = " + context.getRequest().getRequestURI() + " origin = " + origin);
        context.getAttributes().put("originalPath", origin.toExternalForm());
    }
}
