package org.xbib.net.http.template.groovy;

import groovy.text.markup.BaseTemplate;
import org.xbib.net.ParameterException;
import org.xbib.net.http.server.application.BaseApplicationModule;
import org.xbib.net.http.server.application.Application;
import org.xbib.net.http.server.HttpRequest;
import org.xbib.net.http.server.route.HttpRouterContext;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.xbib.net.http.server.service.HttpService;
import org.xbib.settings.Settings;

public class GroovyTemplateApplicationModule extends BaseApplicationModule {

    private static final Logger logger = Logger.getLogger(GroovyTemplateApplicationModule.class.getName());

    private final GroovyMarkupTemplateHandler groovyMarkupTemplateHandler;

    private final GroovyTemplateRenderer groovyTemplateRenderer;

    public GroovyTemplateApplicationModule(Application application, String name) {
        super(application, name);
        ClassLoader classLoader = getClass().getClassLoader();
        Settings settings = Objects.requireNonNull(application.getAttributes().get(Settings.class, "settings"));
        String defaultMarkupTemplate = settings.get("groovy.markup.templateClass",
                "org.xbib.net.http.template.DefaultMarkupTemplate");
        try {
            @SuppressWarnings("unchecked")
            Class<? extends BaseTemplate> defaultMarkupTemplateClass =
                    (Class<? extends BaseTemplate>) Class.forName(defaultMarkupTemplate, true, classLoader);
            logger.log(Level.INFO, "groovy markup template class = " + defaultMarkupTemplateClass.getName());
            this.groovyMarkupTemplateHandler = new GroovyMarkupTemplateHandler(application,
                    classLoader, defaultMarkupTemplateClass, application.getLocale(),
                    settings.getAsBoolean("groovy.markup.autoEscape", true),
                    settings.getAsBoolean("groovy.markup.autoIndent", false),
                    settings.get("groovy.markup.autoIndentString", "  "),
                    settings.getAsBoolean("groovy.markup.autoNewline", false),
                    settings.getAsBoolean("groovy.markup.cacheTemplates", true),
                    settings.get("groovy.markup.declarationEncoding", null),
                    settings.getAsBoolean("groovy.markup.expandEmptyElements", true),
                    settings.get("groovy.markup.newLine", System.getProperty("line.separator")),
                    settings.getAsBoolean("groovy.markup.useDoubleQuotes", true));
            this.groovyTemplateRenderer = new GroovyTemplateRenderer();
        } catch (Exception e) {
            throw new IllegalStateException(e);
        }
    }

    @Override
    public void onOpen(HttpRouterContext httpRouterContext) {
        try {
            groovyMarkupTemplateHandler.handle(httpRouterContext);
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }

    @Override
    public void onOpen(HttpRouterContext httpRouterContext, HttpService httpService, HttpRequest httpRequest) {
        try {
            httpRouterContext.getAttributes().put("request", httpRequest);
            httpRouterContext.getAttributes().put("params", httpRequest.getParameter().asSingleValuedMap());
            application.getModules().forEach(module -> httpRouterContext.getAttributes().put(module.getName(), module));
        } catch (ParameterException e) {
            throw new IllegalStateException(e);
        }
    }

    @Override
    public void onSuccess(HttpRouterContext httpRouterContext) {
        try {
            groovyTemplateRenderer.handle(httpRouterContext);
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }

    @Override
    public void onFail(HttpRouterContext httpRouterContext, Throwable throwable) {
        try {
            groovyTemplateRenderer.handle(httpRouterContext);
        } catch (IOException e) {
            throw new UncheckedIOException(e);
        }
    }
}
